package types

import (
	"errors"
	"fmt"
	"os/exec"
)

type Command struct {
	// Print packets on stderr
	Debug bool

	// Length of packet
	// Larger packets means less packets to exchange for the same amount of data (and less protocol overhead)
	MTU int

	// Values passed in by forward-xxx flags in commandline (forward-xxx:info)
	ForwardInfo map[string][]string

	// List of endpoints the user wants to listen to
	Endpoints []string

	// Map of different sockets provided by user (socket-type flag:socket)
	Sockets map[string]string

	// File where gvproxy's pid is stored
	Pidfile string

	// SSHPort to access the guest VM
	SSHPort int
}

func NewCommand() Command {
	return Command{
		MTU:     1500,
		SSHPort: 2222,
	}
}

// isValidForwardFlag determines whether the flag is a valid forward-xxx flag
func isValidForwardFlag(flag string) bool {
	switch flag {
	case "forward-sock", "forward-dest", "forward-user", "forward-identity":
		return true
	}
	return false
}

// isValidSocketFlag determines whether the flag is a valid listen-xxx flag
func isValidSocketFlag(flag string) bool {
	switch flag {
	case "listen-vpnkit", "listen-qemu", "listen-bess", "listen-stdio", "listen-vfkit":
		return true
	}
	return false
}

// socketsToCmdline converts Command.Sockets to a commandline format
func (c *Command) socketsToCmdline() ([]string, error) {
	args := []string{}

	for socketFlag, socket := range c.Sockets {
		if !isValidSocketFlag(socketFlag) {
			return nil, errors.New("invalid flag: " + socketFlag)
		}
		if socket != "" {
			args = append(args, fmt.Sprintf("-%s %s", socketFlag, socket))
		}
	}

	return args, nil
}

// forwardInfoToCmdline converts Command.ForwardInfo to a command line format
func (c *Command) forwardInfoToCmdline() ([]string, error) {
	args := []string{}

	for forwardInfoFlag, forwardInfo := range c.ForwardInfo {
		if !isValidForwardFlag(forwardInfoFlag) {
			return nil, errors.New("invalid flag: " + forwardInfoFlag)
		}
		for _, i := range forwardInfo {
			if i != "" {
				args = append(args, fmt.Sprintf("-%s %s", forwardInfoFlag, i))
			}
		}
	}

	return args, nil
}

// ToCmdline converts Command to a properly formatted command for gvproxy based
// on its fields
func (c *Command) ToCmdline() ([]string, error) {
	args := []string{}

	// listen (endpoints)
	for _, endpoint := range c.Endpoints {
		args = append(args, "-listen "+endpoint)
	}

	// debug
	if c.Debug {
		args = append(args, "-debug")
	}

	// mtu
	args = append(args, fmt.Sprintf("-mtu %d", c.MTU))

	// ssh-port
	args = append(args, fmt.Sprintf("-ssh-port %d", c.SSHPort))

	// sockets
	socketArgs, err := c.socketsToCmdline()
	if err != nil {
		return nil, err
	}
	args = append(args, socketArgs...)

	// forward info
	forwardInfoArgs, err := c.forwardInfoToCmdline()
	if err != nil {
		return nil, err
	}
	args = append(args, forwardInfoArgs...)

	// pid-file
	if c.Pidfile != "" {
		args = append(args, "-pid-file "+c.Pidfile)
	}

	return args, nil
}

// Cmd converts Command to a commandline format and returns an exec.Cmd which
// can be executed by os/exec
func (c *Command) Cmd(gvproxyPath string) (*exec.Cmd, error) {
	args, err := c.ToCmdline()
	if err != nil {
		return nil, err
	}

	cmd := exec.Command(gvproxyPath, args...)
	return cmd, nil
}
